/*************************************************
* FIPS 186-2 RNG Source File                     *
* (C) 1999-2003 The Botan Project                *
*************************************************/

#include <botan/fips_rng.h>
#include <botan/lookup.h>

namespace Botan {

/*************************************************
* Generate a buffer of random bytes              *
*************************************************/
void FIPS_186_RNG::randomize(byte out[], u32bit length) throw(PRNG_Unseeded)
   {
   if(!is_seeded())
      throw PRNG_Unseeded(name());

   while(length)
      {
      const u32bit copied = std::min(length, buffer.size() - position);
      copy_mem(out, buffer + position, copied);
      out += copied;
      length -= copied;
      update_buffer();
      }
   }

/*************************************************
* Compute the next buffer                        *
*************************************************/
void FIPS_186_RNG::update_buffer() throw()
   {
   buffer = hash->process(xkey);

   for(u32bit j = xkey.size(); j > 0; j--)
      if(++xkey[j-1])
         break;

   byte carry = 0;
   for(u32bit j = xkey.size(); j > 0; j--)
      {
      u16bit sum = (u16bit)xkey[j-1] + buffer[j-1] + carry;
      carry = get_byte(0, sum);
      xkey[j-1] = get_byte(1, sum);
      }
   }

/*************************************************
* Add entropy to internal state                  *
*************************************************/
void FIPS_186_RNG::add_randomness(const byte data[], u32bit length) throw()
   {
   update_entropy(data, length, xkey.size());
   hash->update(xkey);
   hash->update(data, length);
   xkey = hash->final();
   }

/*************************************************
* Check if the the PRNG is seeded                *
*************************************************/
bool FIPS_186_RNG::is_seeded() const
   {
   return (entropy >= 96);
   }

/*************************************************
* Clear memory of sensitive data                 *
*************************************************/
void FIPS_186_RNG::clear() throw()
   {
   hash->clear();
   xkey.clear();
   buffer.clear();
   entropy = position = 0;
   }

/*************************************************
* Return the name of this type                   *
*************************************************/
std::string FIPS_186_RNG::name() const
   {
   return "FIPS-186";
   }

/*************************************************
* FIPS 186-2 RNG Constructor                     *
*************************************************/
FIPS_186_RNG::FIPS_186_RNG()
   {
   hash = get_hash("SHA-1");
   xkey.create(hash->OUTPUT_LENGTH);
   buffer.create(hash->OUTPUT_LENGTH);
   entropy = position = 0;
   }

/*************************************************
* FIPS 186-2 RNG Destructor                      *
*************************************************/
FIPS_186_RNG::~FIPS_186_RNG()
   {
   delete hash;
   }

}
